# -*- coding: utf-8 -*-
from django.core.urlresolvers import reverse


class ProviderMixin:
    def get_provider_url(self, request, resource, name):
        """
        Get provider url given resource and provider name.

        :param request: Django Request
        :param resource:
        :param name:
        :return:
        """
        return request.build_absolute_uri(reverse('health_check:api_{}_{}'.format(resource, name)))



# -*- coding: utf-8 -*-
import json
import sys
from argparse import ArgumentParser
from typing import Dict, Any

import yaml

from health_check.providers import Resource
from health_check.settings import settings


class HealthCheckCommand:
    """
    Show health_check to console.
    """
    CHECK_HEALTH = 'health'
    CHECK_STATS = 'stats'

    OUTPUT_FORMAT_JSON = 'json'
    OUTPUT_FORMAT_YAML = 'yaml'
    OUTPUT_FORMAT_CHOICES = (OUTPUT_FORMAT_JSON, OUTPUT_FORMAT_YAML)

    description = 'HealthCheck command that provides a common entry point for running different checks providers'

    def __init__(self, parse_args=True, stdout=sys.stdout, stderr=sys.stderr):
        """
        Command for running producer, consumer and scheduler processes along with some other utilities.

        :param parse_args: If true, parse sys args.
        :param stdout: Standard output.
        :param stderr: Standard error.
        """
        self.args = self.parse_arguments() if parse_args else {}
        self.stdout = stdout
        self.stderr = stderr

    def add_arguments(self, parser: ArgumentParser, parser_class=None):
        subparsers_kwargs = {}
        if parser_class:
            subparsers_kwargs['parser_class'] = lambda **kwargs: parser_class(self, **kwargs)
        else:
            parser.add_argument('-s', '--settings', help='Settings module')

        subparsers = parser.add_subparsers(title='check', description='HealthCheck checks',
                                           dest='check', **subparsers_kwargs)
        subparsers.required = True

        # Health check
        parser_health = subparsers.add_parser(self.CHECK_HEALTH, help='Health checks')
        parser_health.add_argument(
            '-f', '--output_format', type=str, choices=self.OUTPUT_FORMAT_CHOICES, default=self.OUTPUT_FORMAT_YAML,
            help='Output format'
        )
        parser_health.add_argument(
            '-e', '--error_on_fail', action='store_true', default=False, help='Output format'
        )
        parser_health.set_defaults(func=self.health)

        # Stats
        parser_stats = subparsers.add_parser(self.CHECK_STATS, help='Components stats')
        parser_stats.add_argument(
            '-f', '--output_format', type=str, choices=self.OUTPUT_FORMAT_CHOICES, default=self.OUTPUT_FORMAT_YAML,
            help='Output format'
        )
        parser_stats.set_defaults(func=self.stats)

    def parse_arguments(self) -> Dict[str, Any]:
        """
        Parse sys args and transform it into a dict.

        :return: Arguments parsed in dict form.
        """
        parser = ArgumentParser(description=self.description)
        self.add_arguments(parser)
        return {k: v for k, v in vars(parser.parse_args()).items()}

    def health(self, output_format, *args, **kwargs):
        resource = Resource(kwargs['check'])

        data = resource()

        if output_format == self.OUTPUT_FORMAT_JSON:
            output = json.dumps(data)
        else:
            output = yaml.safe_dump(data, default_flow_style=False)

        self.stdout.write(output)

        if kwargs['error_on_fail'] and not all([value for provider in data.values() for value in provider.values()]):
            sys.exit(1)

    def stats(self, output_format, *args, **kwargs):
        resource = Resource(kwargs['check'])

        if output_format == self.OUTPUT_FORMAT_JSON:
            output = json.dumps(resource())
        else:
            output = yaml.safe_dump(resource(), default_flow_style=False)

        self.stdout.write(output)

    def run(self, **kwargs):
        """
        Command entrypoint.
        """
        kwargs = kwargs or self.args

        if not kwargs:
            raise ValueError('Arguments must be passed or parsed previously')

        if kwargs['settings']:
            settings.build_from_module(module=kwargs['settings'])

        command = kwargs['func']
        return command(**kwargs)



# -*- coding: utf-8 -*-
"""
Built-in Celery check providers.
"""

from health_check.utils import FakeChecker

try:
    from celery.task import control

    celery_inspect = control.inspect()
except ImportError:
    celery_inspect = FakeChecker()


def celery(workers, *args, **kwargs):
    """
    Retrieve the stats data of given celery workers.

    :param workers: List of workers.
    :return: Stats data of each worker.
    """
    try:
        active_workers = celery_inspect.stats() or {}
        workers_stats = {k: v for k, v in active_workers.items() if k in workers}
    except AttributeError:
        workers_stats = None

    return workers_stats



# -*- coding: utf-8 -*-
"""Django application config module.
"""
from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _

from health_check.settings import settings


class HealthCheck(AppConfig):
    name = 'health_check'
    verbose_name = _('Health Check')

    def ready(self):
        settings.build_from_django()



# -*- coding: utf-8 -*-
import os


BASE_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '../'))
APP_PATH = os.path.abspath(os.path.dirname(__file__))

DEBUG = True
TEMPLATE_DEBUG = DEBUG

ADMINS = (
    # ('Your Name', 'your_email@example.com'),
)

MANAGERS = ADMINS

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',  # Add 'postgresql_psycopg2', 'mysql', 'sqlite3' or 'oracle'.
        'NAME': 'test.db',               # Or path to database file if using sqlite3.
        'USER': '',                      # Not used with sqlite3.
        'PASSWORD': '',                  # Not used with sqlite3.
        'HOST': '',                      # Set to empty string for localhost. Not used with sqlite3.
        'PORT': '',                      # Set to empty string for default. Not used with sqlite3.
    },
}

ALLOWED_HOSTS = ['*']

TIME_ZONE = 'Europe/Madrid'

LANGUAGE_CODE = 'en-us'

SITE_ID = 1

USE_I18N = True

USE_L10N = True

USE_TZ = True

MEDIA_ROOT = BASE_DIR + '/media/'

MEDIA_URL = '/media/'

STATIC_ROOT = BASE_DIR + '/static/'

STATIC_URL = '/static/'

STATICFILES_DIRS = (
    APP_PATH + '/static/',
)

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    # 'django.contrib.staticfiles.finders.DefaultStorageFinder',
)

SECRET_KEY = '3+&amp;*6eq%(gs1heqnu!fpb%&amp;#16_^3x#s_nh04^7dxa8tgb%meu'

TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
    # 'django.template.loaders.eggs.Loader',
)

TEMPLATE_CONTEXT_PROCESSORS = (
    'django.contrib.auth.context_processors.auth',
    'django.core.context_processors.debug',
    'django.core.context_processors.i18n',
    'django.core.context_processors.media',
    # "django.core.context_processors.request",
)

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'tests.urls'

WSGI_APPLICATION = 'tests.wsgi.application'

TEMPLATE_DIRS = (
    APP_PATH + '/templates/',
)

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.admin',
    # 'django.contrib.admindocs',
    'health_check',
)



# -*- coding: utf-8 -*-
try:
    from django.core.management.base import BaseCommand, CommandParser
except ImportError:
    BaseCommand = object

from health_check.commands import HealthCheckCommand


class Command(BaseCommand):
    """
    Wrapper that makes a Django command from TaskDispatcherCommand.
    """
    help = HealthCheckCommand.description

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.command = HealthCheckCommand(parse_args=False, stdout=self.stdout, stderr=self.stderr)

    def add_arguments(self, parser):
        self.command.add_arguments(parser=parser, parser_class=CommandParser)

    def handle(self, *args, **options):
        self.command.run(**options)



# -*- coding: utf-8 -*-
"""
Built-in check providers.
"""
import datetime

from git import Repo
from git.exc import InvalidGitRepositoryError

from health_check.settings import settings


def code(*args, **kwargs):
    """
    Code stats, such as current branch, debug mode, last commit, etc.

    :return: Source code stats.
    """
    stats = {'debug': settings.debug}
    try:
        scm_stats = {'scm': 'git'}
        repo = Repo(settings.base_dir)

        # Branch info
        branch = repo.active_branch
        scm_stats['branch'] = {
            'local': branch.name,
        }
        try:
            # Try to get remote branch name
            scm_stats['branch']['remote'] = branch.tracking_branch().name,
        except:
            scm_stats['branch']['remote'] = 'unknown'

        # Last commit info
        commit = repo.commit(branch)
        scm_stats['commit'] = {
            'id': commit.hexsha,
            'summary': commit.summary,
            'author': commit.author.name,
            'date': datetime.datetime.fromtimestamp(commit.authored_date)
        }

        # Get tag info
        tag = None
        tags = repo.tags
        tags.reverse()
        for t in (t for t in tags if tag is None):
            if t.commit == commit:
                tag = t
        if tag:
            scm_stats['tag'] = tag.name
    except InvalidGitRepositoryError:
        scm_stats = {}

    stats.update(scm_stats)

    return stats



# -*- coding: utf-8 -*-
"""
URLs.
"""
from django.conf.urls import include, url
from django.views.generic import TemplateView

app_name = 'health_check'
urlpatterns = [
    url(r'^api/', include('health_check.api.urls')),
    url(r'^$', TemplateView.as_view(template_name='health_check/main.html')),
    url(r'^stats/?$', TemplateView.as_view(template_name='health_check/stats.html')),
    url(r'^health/?$', TemplateView.as_view(template_name='health_check/health.html')),
]



# -*- coding: utf-8 -*-
"""
Views for health_check API.
"""
from django.http import JsonResponse
from django.views.generic import View
from django.views.generic.base import ContextMixin

from health_check.settings import settings
from health_check.api.mixins import ProviderMixin
from health_check.providers.base import Provider, Resource

__all__ = ['ProviderAPIView', 'RootAPIView']


class JSONResponseMixin(object):
    """
    A mixin that can be used to render a JSON response.
    """

    def render_to_json_response(self, context, **response_kwargs):
        """
        Returns a JSON response, transforming 'context' to make the payload.
        """
        return JsonResponse(
            self.get_data(context),
            **response_kwargs
        )

    def get_data(self, context):
        """
        Returns an object that will be serialized as JSON by json.dumps().
        """
        return context or {}


class JSONView(JSONResponseMixin, View):
    """
    View that returns a JSON response.
    """

    def render_to_response(self, context, **response_kwargs):
        return self.render_to_json_response(context, **response_kwargs)


class APIView(JSONView, ContextMixin):
    """
    Base class for API views.
    """

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    def post(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)


class RootAPIView(ProviderMixin, APIView):
    """
    Root API view that routes to each single ProviderAPIView
    """

    def __init__(self):
        super(RootAPIView, self).__init__()
        self.resources = {resource: Resource(resource) for resource in settings.providers}

    def get(self, request, *args, **kwargs):
        """
        Create an object that contains all resources and within them, all providers with their current health_check.

        Mapping of resource to another mapping of provider to health_check.
        Resource -> {Provider -> Status}

        :param request: Request.
        :return: Rendered response.
        """
        context = {resource.name: resource() for resource in self.resources.values()}
        return self.render_to_response(context)

    def options(self, request, *args, **kwargs):
        """
        List all providers of current resource with their absolute urls for each resource

        Mapping of resource to another mapping of provider to url.
        Resource -> {Provider -> URL}

        :param request: Request.
        :return: Rendered response.
        """
        context = {resource.name: {provider_name: self.get_provider_url(request, resource.name, provider_name)
                                   for provider_name in resource.providers.keys()}
                   for resource in self.resources.values()}
        return self.render_to_response(context)


class ResourceAPIView(ProviderMixin, APIView):
    """
    Root API view for a resource. List all checks
    """
    resource = None

    def __init__(self, resource, **kwargs):
        super(ResourceAPIView, self).__init__()
        self.resource = Resource(resource)

    def get_context_data(self, **kwargs):
        return self.resource()

    def options(self, request, *args, **kwargs):
        """
        List all providers of current resource with their absolute urls.

        Mapping of provider to url.
        Provider -> URL

        :param request: Request.
        :return: Rendered response.
        """
        context = {name: self.get_provider_url(request, self.resource.name, name)
                   for name in self.resource.providers.keys()}
        return self.render_to_response(context)


class ProviderAPIView(APIView):
    """
    Specific class that uses a given provider.
    """
    provider = None
    provider_name = None
    provider_args = None
    provider_kwargs = None

    def __init__(self, provider_name, provider, provider_args, provider_kwargs, **kwargs):
        self.provider = Provider(provider_name, provider, provider_args, provider_kwargs)
        super(ProviderAPIView, self).__init__(**kwargs)

    def get_context_data(self):
        return self.provider()



# -*- coding: utf-8 -*-
"""
Settings.
"""
import os
from importlib import import_module

__all__ = ['settings']


class Settings:
    debug = False
    base_dir = None
    installed_apps = ()
    caches = {}
    celery_workers = ()
    providers = {'health': (), 'stats': ()}

    def __init__(self):
        module_path = os.environ.get('HEALTH_CHECK_SETTINGS')
        if module_path:
            self.build_from_module(module_path)

    @staticmethod
    def import_settings(path):
        try:
            s = import_module(path)
        except ImportError:
            m, c = path.rsplit('.', 1)
            module = import_module(m)
            s = getattr(module, c)

        return s

    def _add_celery_providers(self):
        self.providers['health'] += (
            ('celery', 'health_check.providers.celery.health.celery', None, {'workers': self.celery_workers}),
        )
        self.providers['stats'] += (
            ('celery', 'health_check.providers.celery.stats.celery', None, {'workers': self.celery_workers}),
        )

    def build_from_module(self, module=None):
        if isinstance(module, str):
            module = self.import_settings(module)

        # Project path defined in settings
        self.base_dir = getattr(module, 'base_dir', None)

        # Celery application
        self.celery_workers = getattr(module, 'health_check_celery_workers', ())

        # Mapping of resources to list of providers
        # Each provider is a tuple of application name, provider, args, kwargs
        self.providers = getattr(module, 'health_check_providers', {
            'health': (
                ('ping', 'health_check.providers.health.ping', None, None),
            ),
            'stats': (
                ('code', 'health_check.providers.stats.code', None, None),
            )
        })

        # If there is celery workers then add celery providers
        if self.celery_workers:
            self._add_celery_providers()

    def build_from_django(self):
        from django.conf import settings as django_settings

        # Django debug settings.
        self.debug = getattr(django_settings, 'DEBUG', False)

        # Django installed apps
        self.installed_apps = getattr(django_settings, 'INSTALLED_APPS', [])
        self.caches = getattr(django_settings, 'CACHES', {})

        # Project path defined in Django settings
        self.base_dir = getattr(django_settings, 'BASE_DIR', None)

        # Celery application
        self.celery_workers = getattr(django_settings, 'HEALTH_CHECK_CELERY_WORKERS', ())

        # Mapping of resources to list of providers
        # Each provider is a tuple of application name, provider, args, kwargs
        self.providers = getattr(django_settings, 'HEALTH_CHECK_PROVIDERS', {
            'health': (
                ('ping', 'health_check.providers.health.ping', None, None),
                ('databases', 'health_check.providers.django.health.databases', None, None),
                ('caches', 'health_check.providers.django.health.caches', None, None),
            ),
            'stats': (
                ('databases', 'health_check.providers.django.stats.databases', None, None),
                ('code', 'health_check.providers.stats.code', None, None),
            )
        })

        # If there is celery workers then add celery providers
        if self.celery_workers:
            self._add_celery_providers()

settings = Settings()
